      PROGRAM CMP2ARL

C-------------------------------------------------------------------
C LAST REVISION: 09 Feb 1999 (RRD)
C-------------------------------------------------------------------

C Convert Navy COAMPS output files to ARL format compatible for input
C to Hysplit4.  ARL format consists of one file, direct-access I/O, with
C meteorological data ordered by time. COAMPS data will remain on native
C grid and sigma levels.  Grid information is encoded into the ARL format
C index record.  Note that ARL formatted files can be created by time period
C and then "cat" together.  This program will process multiple time periods.
C This program will only extract the minimum fields required by Hysplit.

      LOGICAL   FTEST
      CHARACTER FNAME*80, DATAD*80, BASEF*10

C the following parameter statement sets the maximum grid dimensions to be
C used in subsequent subroutines for both input and output data grids

      PARAMETER (MXP=200, MYP=200, MZP=50)
      PARAMETER (MXYP=MXP*MYP)

C     real data array and packed 2D data array
      REAL RVAR(MXP,MYP,MZP), SVAR(MXP,MYP,MZP), WVAR(MXP,MYP)
      CHARACTER CVAR(MXYP)*1

C     height of each model level and delta-levels
      REAL SIGMA(MZP), DSIGMA(MZP)

C the maximum number of variables supported in ARL format
C where the character field defines the variable type at the surface (0)
C and for the upper levels (1). Units conversion specified for each.

      PARAMETER (MVAR=6)
      CHARACTER*4 VCHAR0(MVAR), VCHAR1(MVAR),
     :            XCHAR0(MVAR), XCHAR1(MVAR)
      REAL        CNVRT0(MVAR), CNVRT1(MVAR)

C     number of ARL variables at the surface and each level
      INTEGER NVAR(0:1)
      DATA    NVAR/6,6/

C     define output variables in ARL character format
      DATA VCHAR0/ 'PRSS','TPPT','UMOF','VMOF','SHTF','SHGT'/ 
      DATA VCHAR1/ 'PRES','UWND','VWND','WWND','TEMP','SPHU'/ 

C     define equivalent COAMPS variables 
      DATA XCHAR0/ 'tepr','prcp','strs','strs','hfls','topo'/ 
      DATA XCHAR1/ 'pppp','uuuu','vvvv','wwww','pott','wvap'/ 

C     units conversion factors for COAMPS to ARL format
      DATA CNVRT0/  1.0, 0.001,  1.0,  1.0,  1.0 , 1.0   /
      DATA CNVRT1/  1.0,   1.0,  1.0,  1.0,  1.0 , 1.0   /

C check for required command line arguments
C where the model initialization time is given on the command line and
C 12 hours of forecast data will be processed with each execution

      NARG=IARGC()
      IF(NARG.LT.4)THEN
         WRITE(*,*)'Usage: cmp2arl [Arg-1] [Arg-2] [Arg-3] [Arg-4]'
         WRITE(*,*)'  1-[Directory] : Data directory with coamps files'
         WRITE(*,*)'  2-[yyyymmddhh]: Time of first file to series'
         WRITE(*,*)'  3-[ngrid]     : Selected coamps grid number'    
         WRITE(*,*)'  4-[hour]      : Starting forecast hour'
         STOP
      END IF

C     decode argument list to determine data processing 

      DO WHILE (NARG.GT.0)
         CALL GETARG(NARG,FNAME)
         IF(NARG.EQ.1)DATAD=FNAME
         IF(NARG.EQ.2)BASEF=FNAME
         IF(NARG.EQ.3)READ(FNAME,'(I1)')NGRID
         IF(NARG.EQ.4)READ(FNAME,'(I1)')NHOUR
         NARG=NARG-1
      END DO

C write ARL packing index file - the file is read by the packing
C subroutine initialization routine (pakset) and the information is
C extracted and written on each time period index record 

C     grid information is decoded from the COAMPS data file   
      CALL HEADER(DATAD,BASEF,NGRID,MZP,MXYP,NSIGMA,MPTYPE,
     :   TNGLAT,ALIGN,NXP,NYP,DELX,XLAT1,XLON1,SIGMA,DSIGMA)
   
      CALL WINDEX(NGRID,NSIGMA,MPTYPE,TNGLAT,ALIGN,NXP,NYP,
     :   DELX,XLAT1,XLON1,SIGMA,MVAR,NVAR,VCHAR0,VCHAR1,FNAME) 
C     compute size of packed data array
      NXYP=NXP*NYP

C     initialize packing subroutines (use same unit number for writing)
      CALL PAKSET(20,FNAME,1,NXP,NYP,(NSIGMA+1))

C construct standardized file name for ARL packed output file and
C open to write 

      FNAME='ARL#_'//BASEF  
      WRITE(FNAME(4:4),'(I1)')NGRID
      LREC=NXYP+50
      OPEN(20,FILE=FNAME,RECL=LREC,ACCESS='DIRECT',FORM='UNFORMATTED')

C the files for each forecast cycle are processed as a single group
C for 12 hours or until there are no more files in the directory

      DO KHOUR=NHOUR,12
C        open and read COAMPS files that correspond to variable selection
         CALL FMTARL(NGRID,NSIGMA,NXP,NYP,NXYP,SIGMA,DSIGMA,MVAR,
     :      NVAR,VCHAR0,VCHAR1,XCHAR0,XCHAR1,CNVRT0,CNVRT1,
     :      DATAD,BASEF,KHOUR,RVAR,CVAR,SVAR,WVAR) 

C        close out time period by writing index record 
         CALL PAKNDX(20)
      END DO
      CLOSE (20)

      END

C-------------------------------------------------------------------

      SUBROUTINE HEADER(DATAD,BASEF,NGRID,MZP,MXYP,NSIGMA,MPTYPE,
     :   TNGLAT,ALIGN,NXP,NYP,DELX,XLAT1,XLON1,SIGMA,DSIGMA)

C decodes COAMPS header file and extracts information required
C for conversion to ARL format and decoding appropriate data files
C Information checked if supported under ARL format conventions

      LOGICAL   FTEST
      REAL      HDATA(500), SIGMA(MZP), DSIGMA(MZP)
      CHARACTER FNAME*80, DATAD*80, BASEF*10

C     construct file name     hhhmmss    level-1  level-2  
      FNAME='dataa#'//BASEF//'0000000'//'00000'//'00000'//'sfl'
C     header file only written for grid #1
      WRITE(FNAME(6:6),'(A1)')'1' 
c     WRITE(FNAME(6:6),'(I1)')NGRID

C     truncate directory string
      KLEN=INDEX(DATAD,' ')-1

C     test for existence of data file before proceeding to decoder
      INQUIRE(FILE=DATAD(1:KLEN)//FNAME,EXIST=FTEST)
      IF(FTEST)THEN
         OPEN(10,FILE=DATAD(1:KLEN)//FNAME,FORM='FORMATTED',
     :         STATUS='OLD',ERR=900)
         WRITE(*,*)'Started processing: ',FNAME
         READ(10,'(5E13.6)')HDATA

C        number of full sigma levels
         NSIGMA=INT(HDATA(2))
         IF(NSIGMA.GT.MZP)THEN
            WRITE(*,*)'ERROR: header - exceeding max allowed levels'
            WRITE(*,*)'  Dimensioned limit  : ',MZP
            WRITE(*,*)'  Levels in data file: ',NSIGMA
            WRITE(*,*)'  Recompile program with larger MZP limits'
            STOP
         END IF    

C        map projection: 1-mercator 2-lambert 3-polar 4-cartesian 5-spherical
         MPTYPE=INT(HDATA(3))
         IF(MPTYPE.GT.3)THEN
            WRITE(*,*)'ERROR: header - map projection not supported'
            WRITE(*,*)'  Projection type: ',MPTYPE
            STOP
         END IF

C        map projection intersection latitudes
         REFLT1=HDATA(4)
         REFLT2=HDATA(5)
C        convert intersection latitude to ARL mapping convention
         TNGLAT=REFLT1
         IF(MPTYPE.EQ.2)TNGLAT=EQVLAT(REFLT1,REFLT2)
C        grid alignment longitude
         ALIGN=HDATA(6)
C        coarse projection standard latitude & longitude 
C        appears to be the center of the grid
         STDLAT=HDATA(7)
         STDLON=HDATA(8)

C        number of grid points in the x and y directions
         NXP=INT(HDATA( (NGRID+2)*10+0 ))
         NYP=INT(HDATA( (NGRID+2)*10+1 ))
         IF(NXP*NXY.GT.MXYP)THEN
            WRITE(*,*)'ERROR: header - exceeding max grid points'
            WRITE(*,*)'  Dimensioned limit : ',MXYP
            WRITE(*,*)'  X,Y points in file: ',NXP,NYP
            WRITE(*,*)'  Recompile program with larger MXP,MYP limits'
            STOP
         END IF    

C        grid spacing (m) in the x and y directions
         DELX=HDATA(9) /3**(NGRID-1)
         DELY=HDATA(10)/3**(NGRID-1)
         IF(DELX.NE.DELY)THEN
            WRITE(*,*)'ERROR: header - only square grid cells supported'
            WRITE(*,*)'   Grid size x,y: ',DELX,DELY
            STOP
         END IF

C        latitude and longitude of 1,1 nest grid point
         XLAT1=HDATA( (NGRID+2)*10+6 )
         XLON1=HDATA( (NGRID+2)*10+7 )

C        enter dsigma level heights into array
         DO K=1,NSIGMA
            DSIGMA(K)=HDATA(100+K)
         END DO

C        enter sigma level heights into array
         DO K=1,NSIGMA
            SIGMA(K)=HDATA(200+K)
         END DO
     
      ELSE
         WRITE(*,*)'ERROR: header - File not found: '
         WRITE(*,*)'  Directory: ',DATAD(1:KLEN)
         WRITE(*,*)'  File     : ',FNAME
         STOP
      END IF
   
C     normal termination  
      CLOSE (10)
      RETURN

C     abnormal termination
  900 WRITE(*,*)'ERROR: header - opening/reading header file: ',FNAME
      STOP
      END

C-------------------------------------------------------------------

      SUBROUTINE WINDEX(NGRID,NSIGMA,MPTYPE,TNGLAT,ALIGN,NXP,NYP,
     :   DELX,XLAT1,XLON1,SIGMA,MVAR,NVAR,VCHAR0,VCHAR1,FNAME) 

C creates the index record configuration file required for ARL packing
C information about grid, variables, and the order they are written.
C The index CFG file is opened by the packing initialization subroutine
C and the file is created only upon the first entry to this routine.
  
C     file name construct
      CHARACTER FNAME*80

C     array to hold grid information
      REAL GRIDS(12)
C     heights of each sigma level
      REAL SIGMA(NSIGMA)
C     number of variables per level
      INTEGER NVAR(0:1)
C     arl variable character IDs for surface and upper level
      CHARACTER*4 VCHAR0(MVAR), VCHAR1(MVAR)

C     grid orientation (meridian parallel to grid)
      GRIDS(6)=0.0

C     lat/lon of grid point 1,1
      GRIDS(8)=1.0
      GRIDS(9)=1.0
      GRIDS(10)=XLAT1 
      GRIDS(11)=XLON1 
C     unused variable 
      GRIDS(12)=0.0

C     defines a polar sterographic projection
      IF(MPTYPE.EQ.3)THEN  

C        pole position of grid projection
         IF(TNGLAT.GT.0.0)THEN
            GRIDS(1)=90.0
         ELSE
            GRIDS(1)=-90.0
         END IF 
         GRIDS(2)=0.0
C        position that defines grid spacing     
         GRIDS(3)=TNGLAT
         GRIDS(4)=ALIGN  
C        delta=x grid size
         GRIDS(5)=DELX/1000.0
C        cone angle   
         GRIDS(7)=90.0

C     defines a merecator projection
      ELSEIF(MPTYPE.EQ.1)THEN 

C        pole position of grid projection
         GRIDS(1)=90.0
         GRIDS(2)=0.0
C        position that defines grid spacing     
         GRIDS(3)=TNGLAT  
         GRIDS(4)=ALIGN
C        longitudinal direction grid length
         GRIDS(5)=DELX/1000.0
C        cone angle   
         GRIDS(7)=0.0

C     defines a lambert conformal projection
      ELSEIF(MPTYPE.EQ.2)THEN

C        pole position of grid projection  
         GRIDS(1)=TNGLAT
         GRIDS(2)=ALIGN
C        position that defines grid spacing     
         GRIDS(3)=TNGLAT
         GRIDS(4)=ALIGN  
C        delta=x grid size
         GRIDS(5)=DELX/1000.0
C        cone angle 
         GRIDS(7)=TNGLAT

      END IF 
    
C configuration file is constructed each time the program is executed
C filename should be unique for each series of packed data
 
      FNAME='CFG#_COAMP'
      WRITE(FNAME(4:4),'(I1)')NGRID

C     write the packer configuration file
      OPEN(10,FILE=FNAME)      

C     write the abbreviated data set name (by grid number)
      WRITE(10,'(20X,A3,I1)')'CMP',NGRID
C     grid number 99 and 3 for terrain coordinate system
      WRITE(10,'(20X,I4)') 99, 3
C     information about grid orientation
      WRITE(10,'(20X,F10.2)')(GRIDS(I),I=1,12)
C     grid size and levels (+1 for surface)
      WRITE(10,'(20X,I4)')NXP,NXP,(NSIGMA+1)

C     surface level information
      HEIGHT=0.0
      WRITE(10,'(20X,F6.0,I3,20(1X,A4))')
     :  HEIGHT,NVAR(0),(VCHAR0(N),N=1,NVAR(0))

C     upper level information
      DO K=1,NSIGMA
         HEIGHT=SIGMA(NSIGMA-K+1) 
         WRITE(10,'(20X,F6.0,I3,20(1X,A4))')
     :     HEIGHT,NVAR(1),(VCHAR1(N),N=1,NVAR(1))
      END DO

      CLOSE (10)
      RETURN
      END

C--------------------------------------------------------------------

      SUBROUTINE FMTARL(NGRID,NSIGMA,NXP,NYP,NXYP,SIGMA,DSIGMA,MVAR,
     :   NVAR,VCHAR0,VCHAR1,XCHAR0,XCHAR1,CNVRT0,CNVRT1,
     :   DATAD,BASEF,KHOUR,RVAR,CVAR,SVAR,WVAR) 

C opens each COAMPS file for a specified time according the equivalent
C variable list of ARL data to pack. Units converted if necessary and 
C written to ARL packed data format, one record per variable per level.

C     file name construct and testing
      CHARACTER FNAME*36, DATAD*80, BASEF*10, BTIME*7
      LOGICAL FTEST

C     heights and delta of each sigma level
      REAL SIGMA(NSIGMA), DSIGMA(NSIGMA)
C     number of variables per level
      INTEGER NVAR(0:1)
C     arl variable character IDs for surface and upper level
      CHARACTER*4 VCHAR0(MVAR), VCHAR1(MVAR)
C     coamps equivalent character IDs for surface and upper level
      CHARACTER*4 XCHAR0(MVAR), XCHAR1(MVAR)
C     coamps to arl unit conversion constants
      REAL        CNVRT0(MVAR), CNVRT1(MVAR)
      
C     unpacked output array
      REAL RVAR(NXP,NYP,NSIGMA), SVAR(NXP,NYP,NSIGMA), WVAR(NXP,NYP)
C     packed output array
      CHARACTER CVAR(NXYP)*1

C file name from variable, date, and time. Note that ARL packed format
C data use 2 digit year identification.  All related Hysplit4 data
C subroutines Y2K compliant.  Two digit year format goes back to data 
C archive starting in 1987.

      KMIN=0
C     extract current date from file name base
      READ(BASEF,'(I4,3I2)')KYR,IMO,IDA,IHR
      IYR=KYR-100*INT(KYR/100)
C     adjust for forecast hour  
      CALL TMPLUS(IYR,IMO,IDA,IHR,KHOUR)                          

C process surface variables first

      DO N=1,NVAR(0)
C        default units conversion
         CNVRT=CNVRT0(N)

C        special processing for variables with no equivalent
         IF(XCHAR0(N).EQ.'topo')THEN
C           special constant fields base time = 0 
            BTIME='0000000'
         ELSEIF(XCHAR0(N).EQ.'prcp')THEN  
C           precip only available at 12 forecast, req each time
            BTIME='0120000'
            CNVRT= FLOAT(KHOUR)/12.0
         ELSE
C          construct base time (hhhmmss) for file name
           BTIME='###0000'
           WRITE(BTIME(1:3),'(I3.3)')KHOUR
           IF(XCHAR0(N).EQ.'strs')THEN  
C             component stress not available, but write as component
C             so that sum of component squared equals scalar stress
              CNVRT=0.7071068
            END IF
         END IF

C        construct file name    
         FNAME=XCHAR0(N)//'a#'//BASEF//BTIME//'00000'//'00000'//'sfl'
C        add current grid number to name
         WRITE(FNAME(6:6),'(I1)')NGRID
C        truncate directory string
         KLEN=INDEX(DATAD,' ')-1

C        test for existence of data file before proceeding to decoder
         INQUIRE(FILE=DATAD(1:KLEN)//FNAME,EXIST=FTEST)
         IF(FTEST)THEN

            OPEN(10,FILE=DATAD(1:KLEN)//FNAME,FORM='UNFORMATTED',
     :          ACCESS='SEQUENTIAL',STATUS='OLD',ERR=900)
            WRITE(*,*)'Started processing: ',FNAME
            READ(10)((RVAR(I,J,1),I=1,NXP),J=1,NYP)
            CLOSE (10)

C           perform any required units conversion
            CALL DATCNV(.FALSE.,RVAR(1,1,1),NXP,NYP,CNVRT,0.0,
     :           RMIN,RMAX)
C           then pack into ARL format and continue
            CALL PAKREC(20,RVAR(1,1,1),CVAR,NXP,NYP,NXYP,VCHAR0(N),
     :           IYR,IMO,IDA,IHR,KMIN,KHOUR,1,0)
         ELSE
            IF(XCHAR0(N).EQ.'tepr')THEN
C              surface pressure variable will be computed from upper level
C              save index for use in upper level loop
               KSFCP=N 
            ELSE
               WRITE(*,*)'WARNING: fmtarl - file not found: ',FNAME
            END IF
         END IF
      END DO

C process data files at sigma levels

      DO N=1,NVAR(1)

C        col:    1->4    56   7->16  17->23  24->28   29->33  34->36
         FNAME=XCHAR1(N)//'a#'//BASEF//BTIME//'00000'//'00000'//'sgl'

C        update name with level information
         WRITE(FNAME(24:28),'(I5.5)')INT(SIGMA(1))
         WRITE(FNAME(29:33),'(I5.5)')INT(SIGMA(NSIGMA))
         IF(XCHAR1(N).EQ.'wwww')THEN
            WRITE(FNAME(24:28),'(I5.5)')
     :            INT(SIGMA(1)+DSIGMA(1)/2.0)
            WRITE(FNAME(29:33),'(I5.5)')
     :            INT(SIGMA(NSIGMA)-DSIGMA(NSIGMA)/2.0)
         END IF

C        add current grid number to name
         WRITE(FNAME(6:6),'(I1)')NGRID
C        truncate directory string
         KLEN=INDEX(DATAD,' ')-1
C        test for existence of data file before proceeding to decoder
         INQUIRE(FILE=DATAD(1:KLEN)//FNAME,EXIST=FTEST)
         IF(FTEST)THEN

            OPEN(10,FILE=DATAD(1:KLEN)//FNAME,FORM='UNFORMATTED',
     :          ACCESS='SEQUENTIAL',STATUS='OLD',ERR=900)
            WRITE(*,*)'Started processing: ',FNAME

C           special processing for vertical velocity due to staggered
C           coordinate system from the remaining diagnostic variables

            IF(XCHAR1(N).EQ.'wwww')THEN
C              vertical velocities defined at level interfaces with
C              additional wvar is the value at the ground surface
C              requiring NSIGMA+1 data levels for input
               READ(10)RVAR,WVAR 
C              average vertical velocities to define mid-point value
               CALL WWWCNV(RVAR,WVAR,NXP,NYP,NSIGMA)

            ELSE
C              all other diagnostic variables of NSIGMA levels
               READ(10)RVAR
            END IF
            CLOSE (10)

C           loop through each level and pack as data record
            DO K=1,NSIGMA

C              perform any required units conversion
               CALL DATCNV(.FALSE.,RVAR(1,1,K),NXP,NYP,
     :              CNVRT1(N),0.0,RMIN,RMAX)

C              shift winds down by 0.5 grid points
               IF(XCHAR1(N).EQ.'uuuu')
     :            CALL UUUCNV(RVAR(1,1,K),WVAR,NXP,NYP)
               IF(XCHAR1(N).EQ.'vvvv')
     :            CALL VVVCNV(RVAR(1,1,K),WVAR,NXP,NYP)

C              convert potential temperature to temperature
               IF(XCHAR1(N).EQ.'pott')
     :            CALL TMPCNV(RVAR(1,1,K),SVAR(1,1,K),NXP,NYP)

C              then pack into ARL format and continue
               CALL PAKREC(20,RVAR(1,1,K),CVAR,NXP,NYP,NXYP,VCHAR1(N),
     :              IYR,IMO,IDA,IHR,KMIN,KHOUR,(NSIGMA-K+2),0)
            END DO

C           additional surface variables estimated from upper levels
            IF(XCHAR1(N).EQ.'pppp')THEN
C              raise surface pressure 1 hPa for every 10 meters
               ADJUST=SIGMA(NSIGMA)/10.0
               CALL DATCNV(.FALSE.,RVAR(1,1,NSIGMA),NXP,NYP,
     :              1.0,ADJUST,RMIN,RMAX)
               CALL PAKREC(20,RVAR(1,1,NSIGMA),CVAR,NXP,NYP,NXYP,
     :              VCHAR0(KSFCP),IYR,IMO,IDA,IHR,KMIN,KHOUR,1,0)
   
C              also save for processing potential temperature
               DO K=1,NSIGMA
               DO J=1,NYP
               DO I=1,NXP
                  SVAR(I,J,K)=RVAR(I,J,K)
               END DO
               END DO
               END DO
            END IF

         ELSE
            WRITE(*,*)'WARNING: fmtarl - file not found: ',FNAME
         END IF
      END DO

      RETURN

  900 WRITE(*,*)'ERROR: fmtarl - opening file: ',FNAME
      STOP
      END

C-----------------------------------------------------------

      SUBROUTINE DATCNV(DIAG,RVAR,NXP,NYP,CNVRT,ADJUST,RMIN,RMAX)

C units conversion from coamps to arl

      LOGICAL DIAG
      REAL RVAR(NXP,NYP)

      RMIN=+1.0E+25
      RMAX=-1.0E-25

      DO J=1,NYP
      DO I=1,NXP
         IF(CNVRT.NE.1.0.OR.ADJUST.NE.0.0)
     :      RVAR(I,J)=RVAR(I,J)*CNVRT+ADJUST
         RMIN=MIN(RMIN,RVAR(I,J))
         RMAX=MAX(RMAX,RVAR(I,J))
      END DO
      END DO

      IF(DIAG)WRITE(*,*)'Minimum: ',RMIN,'    Maximum: ',RMAX

      RETURN
      END

C-----------------------------------------------------------

      SUBROUTINE TMPCNV(RVAR,SVAR,NXP,NYP)

C convert potential temperature to temperature

      REAL RVAR(NXP,NYP), SVAR(NXP,NYP)

      DO J=1,NYP
      DO I=1,NXP
         RVAR(I,J)=RVAR(I,J)*(SVAR(I,J)/1000.0)**0.286
      END DO
      END DO

      RETURN
      END

C-----------------------------------------------------------

      SUBROUTINE WWWCNV(RVAR,WVAR,NXP,NYP,NSIGMA)

C average vertical velocities to match height of other diagnostic
C variables

      REAL RVAR(NXP,NYP,NSIGMA), WVAR(NXP,NYP)

      DO J=1,NYP
      DO I=1,NXP
C        the k index starts from the top and goes down
         DO K=1,(NSIGMA-1)
            RVAR(I,J,K)=0.5*(RVAR(I,J,K)+RVAR(I,J,K+1))
         END DO
C        ground level value defined as additional variable wvar
         RVAR(I,J,NSIGMA)=0.5*(RVAR(I,J,NSIGMA)+WVAR(I,J))
      END DO
      END DO

      RETURN
      END

C------------------------------------------------------------

      SUBROUTINE UUUCNV(RVAR,WVAR,NXP,NYP)

C interpolate u shifted components to position of thermodynamic
C variables (u is offset by +0.5 in x)

      REAL RVAR(NXP,NYP), WVAR(NXP,NYP)

      DO J=1,NYP

C        at the edges special action, upper edge always zero
         WVAR(1,J)=RVAR(1,J)
         WVAR(2,J)=0.5*(RVAR(1,J)+RVAR(2,J))
         WVAR(NXP,J)=RVAR(NXP-1,J)
         WVAR(NXP-1,J)=0.5*(RVAR(NXP-1,J)+RVAR(NXP-2,J))

         DO I=3,(NXP-2)
C           initialize regression coefficient
            CALL LINEAR(0,X,Y,B0,B1,B2)

C           sum preceeding two and forward by one
            DO II=I-2,I+1
               CALL LINEAR(1,FLOAT(II),RVAR(II,J),B0,B1,B2)
            END DO

C           compute regression coefficients
            CALL LINEAR(-1,X,Y,B0,B1,B2)

C           and then compute variable value at thermodynamic point
C           which is shifted down by 0.5 grid points
            XP=I-0.5
            WVAR(I,J)=B2*XP*XP+B1*XP+B0
         END DO
  
C        fill new data back into array
         DO I=1,NXP
            RVAR(I,J)=WVAR(I,J)
         END DO

      END DO

      RETURN
      END

C------------------------------------------------------------

      SUBROUTINE VVVCNV(RVAR,WVAR,NXP,NYP)

C interpolate v shifted components to position of thermodynamic
C variables (v is offset by +0.5 in y)

      REAL RVAR(NXP,NYP), WVAR(NXP,NYP)

      DO I=1,NXP

C        at the edges special action
         WVAR(I,1)=RVAR(I,1)
         WVAR(I,2)=0.5*(RVAR(I,1)+RVAR(I,2))
         WVAR(I,NYP)=RVAR(I,NYP-1)
         WVAR(I,NYP-1)=0.5*(RVAR(I,NYP-1)+RVAR(I,NYP-2))

         DO J=3,(NYP-2)
C           initialize regression coefficient
            CALL LINEAR(0,X,Y,B0,B1,B2)

C           sum preceeding two and forward by one
            DO JJ=J-2,J+1
               CALL LINEAR(1,FLOAT(JJ),RVAR(I,JJ),B0,B1,B2)
            END DO

C           compute regression coefficients
            CALL LINEAR(-1,X,Y,B0,B1,B2)

C           and then compute variable value at thermodynamic point
C           which is shifted down by 0.5 grid points
            YP=J-0.5
            WVAR(I,J)=B2*YP*YP+B1*YP+B0
         END DO

C        fill new data back into array
         DO J=1,NYP
            RVAR(I,J)=WVAR(I,J)
         END DO

      END DO

      RETURN
      END

C------------------------------------------------------------

      SUBROUTINE LINEAR(K,X,Y,B0,B1,B2)

C solution of the equation: Y = B0 + B1 X + B2 X^2
C for regression solutions 

      SAVE A,B,C,D,E,C1,C2,C3

      IF(K.EQ.0)THEN
C        initialize regression sums  
         A=0.0
         B=0.0
         C=0.0
         D=0.0
         E=0.0
         C1=0.0
         C2=0.0
         C3=0.0

      ELSEIF(K.LT.0)THEN
C        computes solution for set of simultaneous linear equations
C        where x=B0; y=B1; z=B2 as defined in the argument list
C        Ax + By + Cz = C1
C        Bx + Cy + Dz = C2
C        Cx + Dy + Ez = C3

         B2=( (C1*B/A-C2)*(B*C/A-D)-(C1*C/A-C3)*(B*B/A-C) ) /
     :      ( (B*C/A-D)**2 - (C*C/A-E)*(B*B/A-C) )

         B1=(C1*B/A-C2)/(B*B/A-C) - B2*(B*C/A-D)/(B*B/A-C)
 
         B0=C1/A - B1*B/A - B2*C/A

      ELSE
C        perform summation
         A=A+1.0
         B=B+X
         C=C+X*X
         D=D+X*X*X
         E=E+X*X*X*X
         C1=C1+Y
         C2=C2+X*Y
         C3=C3+X*X*Y

      END IF

      RETURN
      END
